import os
from flask import Blueprint, request, jsonify
from flamapy.interfaces.python.FLAMAFeatureModel import FLAMAFeatureModel

operations_bp = Blueprint('operations_bp', __name__, url_prefix='/api/v1/operations')

MODEL_FOLDER = './resources/models/'

def _api_call(operation_name:str):
    # Get files
    uploaded_model = request.files['model']
    
    # Check if file is provided
    if uploaded_model.filename != '':

        # Save file
        uploaded_model.save(os.path.join(
            MODEL_FOLDER, uploaded_model.filename))

        fm=FLAMAFeatureModel(os.path.join(
            MODEL_FOLDER, uploaded_model.filename))
        
        operation =getattr(fm,operation_name)

        if(operation_name=='feature_ancestors'):
            result= operation(request.form["feature"])
        elif(operation_name=='valid_product' or operation_name=='valid_configuration' or operation_name=='filter' or operation_name=="commonality"):
            result= operation(request.form["configuration"])
        else:
            result= operation()
        
        # Remove file
        os.remove(os.path.join(MODEL_FOLDER, uploaded_model.filename))
        
        # Return result
        if (result is None):
          return jsonify(error='Not valid result'), 404
        else:
          return jsonify(result)


'''
This is the set of operations within the fm metamodel. 
TODO This could be self generated by means of implementing classes (Operation)
'''
@operations_bp.route('/atomic_sets', methods=['POST'])
def atomic_sets():
    """This endpoint returns the atomics sets of a feature model.
    Note that this is not using sat based solving, thus, there might 
    happens to be more atomic sets to explicitly defined in the model
    ---
    tags:
      - Atomic sets
    parameters:
      - name: model
        in: formData
        type: file
        required: true
    responses:
      200:
        description: A list of lists features representing the atomic sets
        examples:
          result: 5
    """
    return _api_call("atomic_sets")

@operations_bp.route('/average_branching_factor', methods=['POST'])
def average_branching_factor():
    """This endpoint returns the average branching factor of the feature tree.
    ---
    tags:
      - Average Branching factor
    parameters:
      - name: model
        in: formData
        type: file
        required: true
    responses:
      200:
        description: A float representing the average branching factor
        examples:
          result: 5.1
    """
    return _api_call("average_branching_factor")
        
@operations_bp.route('/core_features', methods=['POST'])
def core_features():
    """This endpoint returns the core features within a feature model.
    ---
    tags:
      - Core Features
    parameters:
      - name: model
        in: formData
        type: file
        required: true
    responses:
      200:
        description: A list of core features
        examples:
          result: ['lettuce', 'tomato', 'onion']
    """
    return _api_call("core_features")

@operations_bp.route('/count_leafs', methods=['POST'])
def count_leafs():
    """This endpoint returns the core features within a feature model.
    ---
    tags:
      - Count leafs
    parameters:
      - name: model
        in: formData
        type: file
        required: true
    responses:
      200:
        description: An integer with the number of leafs within a feature model
        examples:
          result: 2
    """
    return _api_call("count_leafs")

@operations_bp.route('/estimated_number_of_products', methods=['POST'])
def estimated_number_of_products():
    """This endpoint returns an approximation of the number of features within a feature model.
    ---
    tags:
      - Estimated number of products
    parameters:
      - name: model
        in: formData
        type: file
        required: true
    responses:
      200:
        description: An integer representing an estimation of the number of products
        examples:
          result: 2324
    """
    return _api_call("estimated_number_of_products")

@operations_bp.route('/feature_ancestors', methods=['POST'])
def feature_ancestors():
    """This endpoint returns the features that are parent from a selected one.
    ---
    tags:
      - Feature Ancestors
    parameters:
      - name: model
        in: formData
        type: file
        required: true
      - name: feature
        in: formData
        type: string
        required: true
    responses:
      200:
        description: A list of feature leafs
        examples:
          result: ['lettuce', 'tomato', 'onion']
    """
    return _api_call("feature_ancestors")

@operations_bp.route('/leaf_features', methods=['POST'])
def leaf_features():
    """This endpoint returns the features that are leafs within a feature model.
    ---
    tags:
      - Leaf Features
    parameters:
      - name: model
        in: formData
        type: file
        required: true
    responses:
      200:
        description: A list of feature leafs
        examples:
          result: ['lettuce', 'tomato', 'onion']
    """
    return _api_call("leaf_features")
    
@operations_bp.route('/max_depth', methods=['POST'])
def max_depth():
    """This endpoint returns the maximum depth of the feature tree.
    ---
    tags:
      - Max depth
    parameters:
      - name: model
        in: formData
        type: file
        required: true
    responses:
      200:
        description: An integer representing the deep of the tree
        examples:
          result: 5
    """
    return _api_call("max_depth")

'''
This is the set of operations within the sat metamodel. 
'''

@operations_bp.route('/commonality', methods=['POST'])
def commonality():
    """This endpoint returns the commonality.
    ---
    tags:
      - Commonality
    parameters:
      - name: model
        in: formData
        type: file
        required: true
      - name: configuration
        in: formData
        type: file
        required: true
    responses:
      200:
        description: An integer representing the deep of the tree
        examples:
          result: 5
    """
    
    return _api_call("commonality")

@operations_bp.route('/dead_features', methods=['POST'])
def dead_features():
    """This endpoint returns the maximum depth of the feature tree.
    ---
    tags:
      - Dead features
    parameters:
      - name: model
        in: formData
        type: file
        required: true
    responses:
      200:
        description: A list of dead features
        examples:
          result: ['lettuce', 'tomato', 'onion']
    """
    return _api_call("dead_features")

@operations_bp.route('/error_detection', methods=['POST'])
def error_detection():
    """This endpoint returns the list of errors within a feature model.
    ---
    tags:
      - Error detection
    parameters:
      - name: model
        in: formData
        type: file
        required: true
    responses:
      200:
        description: A list of errors
        examples:
          result: ['lettuce is a dead feature', 'tomato is a false optional']
    """
    return _api_call("error_detection")

@operations_bp.route('/false_optional_features', methods=['POST'])
def false_optional_features():
    """This endpoint returns the false optional features within a feature model.
    ---
    tags:
      - False optional features
    parameters:
      - name: model
        in: formData
        type: file
        required: true
    responses:
      200:
        description: A list of false optional features
        examples:
          result:  ['lettuce', 'tomato', 'onion']
    """
    return _api_call("false_optional_features")

@operations_bp.route('/filter', methods=['POST'])
def filter():
    """This endpoint returns the products of a models that contains a certain feature.
    ---
    tags:
      - Filter
    parameters:
      - name: model
        in: formData
        type: file
        required: true
      - name: configuration
        in: formData
        type: file
        required: true
    responses:
      200:
        description: A list of products
        examples:
          result: [['lettuce', 'tomato', 'onion'], ['lettuce2', 'tomato2', 'onion']]
    """
    return _api_call("filter")

@operations_bp.route('/products_number', methods=['POST'])
def products_number():
    """This endpoint returns the number of products within a model.
    ---
    tags:
      - Number of products
    parameters:
      - name: model
        in: formData
        type: file
        required: true
    responses:
      200:
        description: An integer representing the number of products
        examples:
          result: 5
    """
    return _api_call("products_number")

@operations_bp.route('/products', methods=['POST'])
def products():
    """This endpoint returns the list of products of a model.
    ---
    tags:
      - Products
    parameters:
      - name: model
        in: formData
        type: file
        required: true
    responses:
      200:
        description: A list of products
        examples:
          result: [['lettuce', 'tomato', 'onion'], ['lettuce2', 'tomato2', 'onion']]
    """
    return _api_call("products")

@operations_bp.route('/valid_configuration', methods=['POST'])
def valid_configuration():
    """This endpoint returns a boolean indicating if a partial configuration is valid.
    ---
    tags:
      - Valid partial configuration
    parameters:
      - name: model
        in: formData
        type: file
        required: true
      - name: configuration
        in: formData
        type: file
        required: true
    responses:
      200:
        description: A boolean indicating if the configuration is valid
        examples:
          result: True
    """
    return _api_call("valid_configuration")

@operations_bp.route('/valid_product', methods=['POST'])
def valid_product():
    """This endpoint returns a boolean if a fulll configuration (a.k.a product) is valid.
    ---
    tags:
      - Valid full configuration (a.k.a. product)
    parameters:
      - name: model
        in: formData
        type: file
        required: true
      - name: configuration
        in: formData
        type: file
        required: true
    responses:
      200:
        description: A boolean indicating if the configuration is valid
        examples:
          result: True
    """
    return _api_call("valid_product")

@operations_bp.route('/valid', methods=['POST'])
def valid():
    """This endpoint returns a boolean indicating if the model is valid.
    ---
    tags:
      - Valid
    parameters:
      - name: model
        in: formData
        type: file
        required: true
    responses:
      200:
        description: A boolean indicating if the model is valid
        examples:
          result: True
    """
    return _api_call("valid")